#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from setuptools import setup, find_packages

__updated__ = "2017-03-09"
__author__ = "Aurélien Moreau"
__copyright__ = "Copyright 2017, Angus.ai"
__credits__ = ["Aurélien Moreau"]
__license__ = "Apache v2.0"
__maintainer__ = "Aurélien Moreau"
__status__ = "Production"

__version__ = "1.0.4"

setup(name='angus-web-visu',
      version=__version__,
      description='Angus Visualization Toolkit',
      author=__author__,
      author_email='aurelien.moreau@angus.ai',
      url='http://www.angus.ai/',
      install_requires=[
          "angus-sdk-python",
          "tornado",
          "pytz",
          "numpy",
      ],
      packages=find_packages(),
      license=__license__,
      )



# -*- coding: utf-8 -*-

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import StringIO
import datetime
import numpy as np
import logging
import Queue
import time
import multiprocessing

import pytz
import cv2

import angus

LOGGER = logging.getLogger(__name__)

CONF = {
    "appearance": 0.8,
    "disappearance": 0.8,
    "age_estimated": 0.5,
    "gender_estimated": 0.5,
    "focus_locked": 0.9,
    "emotion_detected": 0.4,
    "direction_estimated": 0.8
}

class Worker(multiprocessing.Process):
    def __init__(self, *args, **kwargs):
        self.index = kwargs.pop("index", 0)
        super(Worker, self).__init__(*args, **kwargs)
        self.output = multiprocessing.Queue(20)
        self.capture = None
        self.daemon = True

    def init_capture(self):
        capture = cv2.VideoCapture(self.index)

        capture.set(cv2.cv.CV_CAP_PROP_FRAME_WIDTH, 640)
        capture.set(cv2.cv.CV_CAP_PROP_FRAME_HEIGHT, 480)
        capture.set(cv2.cv.CV_CAP_PROP_FPS, 10)

        LOGGER.info("Capture settings: %sx%s %sfps",
                    capture.get(cv2.cv.CV_CAP_PROP_FRAME_WIDTH),
                    capture.get(cv2.cv.CV_CAP_PROP_FRAME_HEIGHT),
                    capture.get(cv2.cv.CV_CAP_PROP_FPS)
        )

        LOGGER.info("Capture configured")

        if not capture.isOpened():
            LOGGER.error("Capture can not be openened")
            return False

        self.capture = capture
        return True

    def init_angus(self):
        # Try max 5 times angus-server connexion
        conn = angus.connect()
        service = None
        for count in range(5):
            try:
                service = conn.services.get_service("scene_analysis", version=1)
                service.enable_session()
                break
            except Exception as e:
                LOGGER.error(e)
                LOGGER.warning("Scene Analysis not ready (attempt #%s/5), wait 1s and try again",
                               count+1)
                time.sleep(2)
                continue
        if service is None:
            LOGGER.error("Scene analysis is not available, shutdown")
            return False

        LOGGER.info("Scene Analysis service connected")
        self.service = service
        return True

    def loop(self):
        ret, frame = self.capture.read()
        if not ret:
            LOGGER.warning("No image anymore")
            return False

        stamp = datetime.datetime.now(pytz.utc)
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        ret, buff = cv2.imencode(".jpg", gray, [cv2.IMWRITE_JPEG_QUALITY, 80])
        buff = StringIO.StringIO(np.array(buff).tostring())

        try:
            job = self.service.process({"image": buff,
                                    "timestamp" : stamp.isoformat(),
                                    "camera_position": "facing",
                                    "sensitivity": CONF,
            })
        except Exception as e:
            LOGGER.error(e.message)
            LOGGER.error("Unknow error (to be checked), just continue")
            return

        res = job.result
        notifications = []

        if not "error" in res:
            # This parses the events
            if "events" in res:
                for event in res["events"]:
                    value = res["entities"][event["entity_id"]][event["key"]]
                    event_stg = "{} - {}, {}, {}".format(event["entity_id"],
                                                           event["type"],
                                                           event["key"],
                                                           value)
                    notifications.append(event_stg)

            # This parses the entities data
            for key, val in res["entities"].iteritems():
                x, y, dx, dy = map(int, val["face_roi"])
                cv2.rectangle(frame, (x, y), (x+dx, y+dy), (0, 255, 0), 2)

        ret, frame = cv2.imencode(".jpg", frame,  [cv2.IMWRITE_JPEG_QUALITY, 80])
        frame = frame.tostring()

        try:
            self.output.put_nowait((frame, notifications))
        except Queue.Full:
            LOGGER.error("Queue is full")
            return False

        return True

    def run(self):

        if not (self.init_capture() and self.init_angus()):
            self.capture.release()
            return

        LOGGER.info("Start grab camera")

        while self.capture.isOpened():
            if not self.loop():
                break

        LOGGER.info("Release capture")
        self.capture.release()
        LOGGER.info("Image capture down")



# -*- coding: utf-8 -*-

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Create a complete server with:
* 1 worker
* 1 mjpeg server
* notification system
"""
import logging
import os.path
import os
import Queue

import tornado.web
import tornado.ioloop
import tornado.gen

from notifier import EventSource, NotificationPusher
from mjpegserver import MJPEGHandler
from worker import Worker

LOGGER = logging.getLogger(__name__)
PWD = os.path.dirname(__file__)

def unfold(queue, config):
    try:
        new_frame, notifs = queue.get_nowait()
        if len(notifs) > 0:
            config["event_source"].publish(notifs)
        config["last_frame"] = new_frame
    except Queue.Empty:
        pass

def watch_process(process):
    if not process.is_alive():
        LOGGER.warning("Process is down, shutdown")
        tornado.ioloop.IOLoop.instance().stop()

def make_app():
    LOGGER.info("Make server application")

    video_index = 0

    event_source = EventSource()
    config = {
        "last_frame": None,
        "event_source": event_source,
    }

    # Worker
    worker = Worker(index=video_index)
    worker.start()

    # Unfold results
    unfolder = tornado.ioloop.PeriodicCallback(lambda: unfold(worker.output, config), 10)
    unfolder.start()

    # Worker watchdog
    worker_watch = tornado.ioloop.PeriodicCallback(lambda: watch_process(worker), 1000)
    worker_watch.start()

    # Sender
    return tornado.web.Application([
        (r"/mjpeg", MJPEGHandler, dict(config=config)),
        (r"/notifications", NotificationPusher, dict(event_source=event_source)),
        (r"/", tornado.web.RedirectHandler, dict(url="/index.html")),
        (r"/(.*)", tornado.web.StaticFileHandler, {"path": PWD}),
    ])

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    make_app().listen(8888)
    LOGGER.info("Start ioLoop")
    tornado.ioloop.IOLoop.current().start()
